#!/usr/bin/env bash

##############################################################################################
#                          R workflow Utils Functions                                        #
#                                                                                            #
# Functions to facilitate renku workflow creation for workflows that call R scripts          #
# Most functions are generalizable, but some specifically target omnibenchmark scripts and   #
# are only valid or useful in an omnibenchmark context/repository                            #
#                                                                                            #
# Change History                                                                             #
# 10/06/2021  Almut Lütge    Original code: Bundle R workflow functions partly written by    #
#                            Oksana Riba Grognuz and Almut Lütge in this utils.              #
#                                                                                            #
#                                                                                            #
# authors: Oksana Riba Grognuz & Almut Lütge                                                 #
##############################################################################################
##############################################################################################
##############################################################################################

##############################################################################################

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Prerequisites  ----------------------------------------##
## --------------------------------------------------------------------------------------- ##


{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Global variables  -------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/global_vars.sh
} || { 
    echo "Warning, src/utils/global_vars.sh does not exist and may be required for some functions."
}


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##

help_parse_r_arguments()
{
   echo "Function to parse R arguments to R within a renku run call."
   echo
   echo "Syntax: parse_r_arguments R_ARR"
   echo
   echo "Arguments:"
   echo "R_ARR          Array with all R arguments that need to be parsed as R inputs."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_parse_rscript_arguments()
{
   echo "Function to parse R arguments to Rscript within a renku run call."
   echo
   echo "Syntax: parse_rscript_arguments R_ARR"
   echo
   echo "Arguments:"
   echo "R_ARR          Array with all R arguments that need to be parsed as R inputs."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_parse_r_inputs()
{
   echo "Function to print renku input from array with R arguments and input array keys defined in config.sh."
   echo
   echo "Syntax: parse_r_inputs R_ARR"
   echo
   echo "Arguments:"
   echo "R_ARR          Noame of Array with R arguments that shall be defined as renku inputs."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_parse_parameter()
{
   echo "Function to print Prints renku parameter with their defaults as values for renku workflow loop/renku map. Not R script specific."
   echo
   echo "Syntax: parse_parameter PARAM_ARR"
   echo
   echo "Arguments:"
   echo "PARAM_ARR          Name of parameter array with keys as parameter and values as their defaults. Usually defined in config.sh."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_parse_default_params_to_Rargs()
{
   echo "Function to parse parameter and their values to the R argument array (and so R)"
   echo
   echo "Syntax: parse_default_params_to_Rargs RARGS"
   echo
   echo "Arguments:"
   echo "RARGS          Array with R arguments that are also renku inputs."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}



##############################################################################################


## ---------------------------------------------------------------------------------------- ##
## -------------- Functions to define inputs of renku workflows calling R   ----------------##
## ---------------------------------------------------------------------------------------- ##


function parse_r_arguments(){

    # Input: assoicative array with R arguments-values
    # Output: string for R CMD
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_parse_r_arguments
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    # parse arguments
    local -n arr=$1
    local quoted_args
    quoted_args=$(for i in "${!arr[@]}"; do   printf "%s=\"%s\"\n" "${i}" "${arr[$i]}"; done)
    printf %b "--args $quoted_args"
}


function parse_rscript_arguments(){

    # Input: assoicative array with R arguments-values
    # Output: string for R CMD
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_parse_rscript_arguments
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    # parse arguments
    local -n arr=$1
    local quoted_args
    quoted_args=$(for i in "${!arr[@]}"; do   printf "%s=%s " "--${i}" "${arr[$i]}"; done)
    printf %b "$quoted_args"
}


function parse_r_inputs(){

    # Input: Name of assoicative array with R arguments-values
    # Output: Prints renku input from array with R arguments and input array keys defined in config.sh
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_parse_r_inputs
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local -n arr=$1
    local r_input
    r_input=$(echo "${!arr[@]}" "${!IN_PREFIX[@]}" | tr ' ' '\n' | sort | uniq -d)
    
    for i in ${r_input}
    do
       echo "--input ${arr[$i]}"
    done | tr '\n' ' '
}


function parse_parameter(){
    
    # Input: Name of parameter array with keys as parameter and values as their defaults. Usually defined in config.sh.
    # Output: Prints renku parameter with their defaults as values for renku workflow loop/renku map.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_parse_parameter
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    # parse parameter
    local -n arr=${1:-PARAM_AR}
    
    for i in "${!arr[@]}"
    do
       echo "--${i} ${arr[$i]}"
    done | tr '\n' ' '
}



function parse_default_params_to_Rargs(){

    # Input1: Name of assoicative array with R arguments-values of general workflow inputs.
    # Input2: Name of parameter array specifying the method's valid paramerter and their corresponding defaults.  
    # Output: Assoicative array with R arguments-values of inputs and parameter defaults.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_parse_default_params_to_Rargs
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    local -n arr=${1:-R_args}
    local -n param_arr=${2:-PARAM_AR}
    for param in "${!param_arr[@]}"; do
        arr["$param"]="${param_arr[$param]}"
    done

}


##############################################################################################