#!/usr/bin/env bash

##############################################################################################
#                          Dataset Check Functions                                           #
#                                                                                            #
# Functions to perform dataset checks and check dataset function inputs to ensure actions on #
# the targeted datasets.                                                                     #
# Most functions are generalizable, but some specifically target omnibenchmark scripts and   #
# are only valid or useful in an omnibenchmark context/repository                            #
#                                                                                            #
# Change History                                                                             #
# 05/07/2021  Almut Lütge         Initialize check functions and generating setup.           #
#                                                                                            #
# Authors: Almut Lütge                                                                       #
##############################################################################################
##############################################################################################
##############################################################################################

##############################################################################################

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Prerequisites  ----------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}


# Error function
error_exit()
{
  printf '%s\n' "$@" 1>&2
  exit 1
} 


## --------------------------------------------------------------------------------------- ##
## -------------------------------- Global variables  -------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/global_vars.sh
} || { 
    echo "Warning, src/utils/global_vars.sh does not exist and may be required for some functions."
}





##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##


help_check_import_from_target()
{
   echo "Function to check dataset ids to point to projects, where their files were generated as part of renku workflows."
   echo "In particular this function checks if there is an edge with the dataset files as target in the project that is linked to a dataset id."
   echo "By default this function will only check one randomly selected file from the dataset, but not the meta file."
   echo
   echo "Syntax: check_import_from_target DATASET_IDS"
   echo
   echo "Arguments:"
   echo "DATASET_IDS    Dataset_ids to check for file target edges."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo "a              Check if any of the first -a dataset files is associated with a target edge."
   echo
}


help_check_project_status()
{
   echo "Function to check dataset ids to point to projects with successful pipelines."
   echo "In particular this function checks the orchestrator cicd pipeline status of the project, that a dataset id originates from."
   echo
   echo "Syntax: check_project_status DATASET_IDS"
   echo
   echo "Arguments:"
   echo "DATASET_IDS    Dataset_ids to check for the project status for."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo "o              Orchestrator namespace without gitlab prefix, e.g. defaults to omnibenchmark/orchestrator"
   echo
}


help_get_status_by_path()
{
   echo "Function to get the CICD status of a specified project."
   echo "In particular this function checks the orchestrator cicd pipeline status of the last included - not skipped - run."
   echo
   echo "Syntax: get_status_by_path PROJECT_PATH"
   echo
   echo "Arguments:"
   echo "PROJECT_PATH   Path to the project."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo "o              Orchestrator namespace without gitlab prefix, e.g. defaults to omnibenchmark/orchestrator"
   echo
}


help_check_orchestrator()
{
   echo "Function to check dataset ids to point to projects that are part of an omnibenchmark orchestrator."
   echo "In particular this function checks the project, that a dataset id originates from has been run in a SUCCESSFUL omnibenchmark pipeline before"
   echo
   echo "Syntax: check_orchestrator -o ORCHESTRATOR_PATH DATASET_IDS"
   echo
   echo "Arguments:"
   echo "DATASET_IDS    Dataset_ids to check for the project status for."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo "o              Orchestrator namespace without gitlab prefix, e.g. defaults to omnibenchmark/orchestrator"
   echo
}


help_find_last_included_pipeline_id()
{
   echo "Function to get the last orchestrator pipeline id of a specified project path"
   echo "In particular this function returns the most recent orchestrator pipeline id in which this project path was not skipped."
   echo
   echo "Syntax: find_last_included_pipeline_id -o ORCHESTRATOR_PATH PROJECT_PATH"
   echo
   echo "Arguments:"
   echo "PROJECT_PATH   Path to the project to find the pipeline id for."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo "o              Orchestrator namespace without gitlab prefix, e.g. defaults to omnibenchmark/orchestrator"
   echo
}


help_is_part_of_orchestrator()
{
   echo "Function to check if a project path is part of a orchestrator pipeline."
   echo "In particular this function checks the last successful orchestrator cicd pipeline for having this project included."
   echo
   echo "Syntax: is_part_of_orchestrator ORCHESTRATOR_ID PROJECT_PATH"
   echo
   echo "Arguments:"
   echo "ORCHESTRATOR_ID   Gitlab project ID of the orchestrator project to check."
   echo "PROJECT_PATH      Path to the project to check."
   echo
   echo "Options:"
   echo "h                 Print this Help."
   echo
}


help_orchestrator_projects_by_id()
{
   echo "Function to get all orchestrator pipelines projects."
   echo "In particular this function returns all project urls from the last successful orchestrator cicd pipeline."
   echo
   echo "Syntax: orchestrator_projects_by_id ORCHESTRATOR_ID PROJECT_PATH"
   echo
   echo "Arguments:"
   echo "ORCHESTRATOR_ID   Gitlab project ID of the orchestrator project to check."
   echo "PROJECT_PATH      Path to the project to check."
   echo
   echo "Options:"
   echo "h                 Print this Help."
   echo
}



##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Helper functions/snippets -------------------------------##
## ---------------------------------------------------------------------------------------- ##

function graphql_query () {
    
    # Input1: Path to file to query
    # Input2: Path to project to perform query in
    # Output: Nodes and edges for a specific graphql endpoint
    
    local file_path=$1
    local project_path=$2

    curl -X POST -v -H "Content-Type: application/json" "$GRAPHQL_URL" -d '{ "query": "{ lineage(projectPath: \"'"${project_path}"'\", filePath: \"'"${file_path}"'\") { nodes { id label } edges { source target } } }"}'
    
}

# The same as in dataset_utils, but can't source them in each other 
function dataset_files_by_id () {

   local dataset_id=$1
   curl -s "${KG_URL}/datasets/${dataset_id}" | jq -r ".hasPart | .[] | .atLocation"
   }



function get_gitlab_id_by_path () {

   local project_path=$1
   git_id=$(curl -s "${KG_URL}/projects/$project_path"| jq -r '.identifier')
   
   if [[ "$git_id" =~ null ]]
   then
        error_exit "ERROR: Path not found. Please check that $project_path exists."
   fi
   
   echo "$git_id"
   
   }

##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ----------------------- Functions to check dataset import by keywords -------------------##
## ---------------------------------------------------------------------------------------- ##

function orchestrator_projects_by_id () {

   # Input1:  Orchestrator ID
   # Output:  project_paths that are included in the orchestrator 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_orchestrator_projects_by_id
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))

   local o_id=$1
   local status="skipped"
   local pipe_num=1
   
   pipe_ids=$(curl "${GIT_API}/projects/${o_id}/pipelines?status=success" | jq -r '.[] | .id')
   pipe_len=$(echo $pipe_ids | wc -w)
           
    
  #loop through pipelines until project status is different from skipped
    while [[ $status =~ skipped ]] && [[ $pipe_num -le $pipe_len ]]
    do
        pipe_id=$(echo $pipe_ids | awk -v a=$pipe_num '{print $a}')
        skip=$(curl "${GIT_API}/projects/${o_id}/pipelines/${pipe_id}/bridges?scope=skipped" | jq -r '.[]')
                   
        if [[ -z $skip ]]
        then
            status="in"
        else
            status="skipped"
            pipe_num=$(($pipe_num+1))
        fi
        
    done
   
   if [[ "$status" == "skipped" ]]
   then
       error_exit "ERROR: No complete orchestrator run found. All successful runs contain skipped trigger." \
                  "Please make sure that project and orchestrator id are correct:" \
                  "Orchestrator ID: $o_id; Project path: $project_path"
   fi
   
   
   o_urls=$(curl "${GIT_API}/projects/${o_id}/pipelines/${pipe_id}/bridges" | \
            jq -r '.[] | .downstream_pipeline | .web_url')
   
   echo "$o_urls"
   
   }


   
function is_part_of_orchestrator () {

   # Input1:  Orchestrator ID
   # Input2:  Project path
   # Output:  orchestrator url of input project or empty, if project path is not part of orchestrator 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_is_part_of_orchestrator
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))

   local o_id=$1
   local project_path=$2
   o_urls=$(orchestrator_projects_by_id "$o_id")
       
   overlap=$(printf '%s\n' "${o_urls}" | grep -E "$project_path")
   
   echo "$overlap"
   
   }
   

function find_last_included_pipeline_id () {
    
   # Input:  Project path to find orchestrator pipeline id for 
   # Output: Pipeline id of the orchestrator that had this projects pipeline included and not skipped
       
    local OPTIND o_path="omnibenchmark/orchestrator"
    
    # Display help message
    while getopts ":o:h" option; do
        case $option in
            h) 
                help_find_last_included_pipeline_id
                return 1;;
            o)
                o_path=${OPTARG};;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   local project_path=$1
   local o_id=$(get_gitlab_id_by_path "$o_path")
   local status="skipped"
   local pipe_num=0
   
   #check if project path is part of orchestrator
   is_part=$(is_part_of_orchestrator "$o_id" "$project_path")
   
   if [[ -z ${is_part} ]]
   then
       error_exit "ERROR: Project is not part of the orchestrator defined." \
                  "Please make sure that project and orchestrator id are correct:" \
                  "Orchestrator ID: $o_id; Project path: $project_path"
   fi
   
   #loop through pipelines until project status is different from skipped
    while [ "$status" == "skipped" ]
    do
        pipe_id=$(curl "${GIT_API}/projects/${o_id}/pipelines" | jq -r '.['$pipe_num'] | .id')
        included=$(curl "${GIT_API}/projects/${o_id}/pipelines/${pipe_id}/bridges" | \
                   jq -r '.[] | .downstream_pipeline | .web_url' | grep "$project_path")
                   
        if [[ -z $included ]]
        then
            status="skipped"
            pipe_num=$(($pipe_num+1))
        else
            status="in"
        fi
    done
   
   echo "${pipe_id}"
}



function check_import_from_target () {
    
   # Input:  Name of dataset IDs to import 
   # Output: Dataset IDs from datasets that have been imported, but are not the original dataset
    local OPTIND any=""
    
    # Display help message
    while getopts ":a:h" option; do
        case $option in
            h) 
                help_check_import_from_target
                return 1;;
            a)
                any=${OPTARG};;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   dataset_ids=$*
   artificial_ids=()
   
   # select data creation repository as source
   for id in ${dataset_ids}
   do
       #check that id is the most recent in it's project
       project_path=$(curl -s "${KG_URL}/datasets/$id" | jq -r '.'\"project\"' | .'\"path\")
       
       latest=$(curl -s "${KG_URL}/projects/${project_path}/datasets" | jq -r '.[].'\"identifier\")
       
       if [[ ! "${latest}" =~ ${id} ]]
       then
           artificial_ids+=("$id")
       else             
           # Check that files have target edges
           if [[ -n $any ]]
           then
               test_files=$(dataset_files_by_id "$id" | head -"${any}")
               n_files=$(echo $test_files | awk '{print NF}')
               i=1
               wflow_edge=false
           
               while (( $wflow_edge == "false" && $i <= $n_files ))
               do
                   file=$(echo $test_files | awk -v a=$i '{print $a}')
                                 
                   target_edges=$(graphql_query "$file" "$project_path" | \
                                    jq '.data? | .lineage? | .edges[]? | select(.target | contains('\"${file}\"')) | length' | wc -l)
           
                   if [[ $target_edges -gt 0 ]]
                   then
                       wflow_edge=true
                   fi
               
                   i=$(( $i + 1 ))
               
               done
           
               if [[ $wflow_edge == "false" ]]
               then
                   artificial_ids+=("$id")
               fi               
            else
               if [[ -v IN_PREFIX[@] && -n "${IN_PREFIX['meta_file']}" ]]
               then
                   test_files=$(dataset_files_by_id "$id" | grep -v "${IN_PREFIX['meta_file']}" | head -1)
               else
                   test_files=$(dataset_files_by_id "$id" | head -1)
               fi

               if [[ -z $test_files ]]
               then
                   artificial_ids+=("$id")
               fi
       
               # find out where test files were generated
               art_ids=()
       
               for file in ${test_files}
               do                      
                   target_edges=$(graphql_query "$file" "$project_path" | \
                                    jq '.data? | .lineage? | .edges[]? | select(.target | contains('\"${file}\"')) | length' | wc -l)
           
                   if [[ $target_edges -eq 0 ]]
                   then
                       art_ids+=("$id")
                   fi    
               done
       
               art_length=$(printf '%s\n' "${art_ids[@]}" | wc -w)
       
               if [[ ${art_length} -gt 0 ]]
               then
                   artificial_ids+=("$id")
               fi
            fi
       fi
   done
    
   origin_ids=$(printf '%s\n' ${dataset_ids[@]} ${artificial_ids[@]} | sort | uniq -u)
   
   # Test output
   if [[ -z ${origin_ids} ]]
   then
       error_exit "ERROR: No datasets from projects with their file workflows found." \
                  "If you are sure that you want to import datasets anyways, please check the dataset id and do so explicitly using:" \
                  "renku dataset import -y 'DATASET_URL'"
   fi
 
   echo "${origin_ids[@]}"
}




function get_status_by_path () {
    
   # Input:  Path to project to get the status for 
   # Output: CICD status of the last pipeline, that was included into the orchestrator and not skipped
   
   local OPTIND o_path="omnibenchmark/orchestrator"
       
    # Display help message
    while getopts ":o:h" option; do
        case $option in
            h) 
                help_get_status_by_path
                return 1;;
            o)
                o_path=${OPTARG};;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   project_path=$1
   o_id=$(get_gitlab_id_by_path "$o_path")
   project_id=$(get_gitlab_id_by_path "$project_path")
   latest_pipe=$(find_last_included_pipeline_id -o "$o_path" "$project_path")
                    
   pipe_status=$(curl "${GIT_API}/projects/${o_id}/pipelines/${latest_pipe}/bridges" | jq -r '.[] | .downstream_pipeline | select(.project_id != null) | select(.project_id | contains('"$project_id"')) | .status')
       
   echo "${pipe_status}"
}



function check_project_status () {
    
   # Input:  Name of dataset IDs to import 
   # Output: Dataset IDs from projects with cicd status success
   
   local OPTIND o_path="omnibenchmark/orchestrator"
       
    # Display help message
    while getopts ":o:h" option; do
        case $option in
            h) 
                help_project_status
                return 1;;
            o)
                o_path=${OPTARG};;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   dataset_ids=$*
   failed_ids=()
   
   # get pipeline status via gitlab API
   for id in ${dataset_ids[@]}
   do
       project_path=$(curl -s "${KG_URL}/datasets/$id" | jq -r '.'\"project\"' | .'\"path\")                    
       pipe_status=$(get_status_by_path -o "$o_path" "$project_path")
       
       if [[ ! "${pipe_status}" =~ success ]]
       then
           failed_ids+=("$id")
       fi             
    done
    
   success_ids=$(printf '%s\n' ${dataset_ids[@]} ${failed_ids[@]} | sort | uniq -u)
   
   # Test output
   if [[ -z ${success_ids} ]]
   then
       error_exit "ERROR: No datasets with successful project status found." \
                  "Please check the imported projects for failures and fix them."
   fi
 
   echo "${success_ids[@]}"
}




function check_orchestrator () {
    
   # Input:  Name of dataset IDs to import 
   # Output: Dataset IDs from datasets that are part of the ormnibenchmark orchestrator
   
   local OPTIND o_path="omnibenchmark/orchestrator"
    
    # Display help message
    while getopts ":o:h" option; do
        case $option in
            h) 
                help_check_orchestrator
                return 1;;
            o)
                o_path=${OPTARG};;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   dataset_ids=$*
   excluded_ids=()
   
   # select data creation repository as source
   for id in ${dataset_ids}
   do
       # datasets project path
       project_path=$(curl -s "${KG_URL}/datasets/$id" | jq -r '.'\"project\"' | .'\"path\")
       
       # orchestrator id and pipeline _id of the last successful(!) pipeline
       o_id=$(get_gitlab_id_by_path "$o_path")      
       overlap=$(is_part_of_orchestrator "$o_id" "$project_path")
       
       if [[ -z $overlap ]]
       then
           excluded_ids+=("$id")
       fi             
    done
    
   success_ids=$(printf '%s\n' ${dataset_ids[@]} ${excluded_ids[@]} | sort | uniq -u)
   
   # Test output
   if [[ -z ${success_ids} ]]
   then
       error_exit "ERROR: No datasets associated to a SUCCESSFUL omnibenchmark orchestrator pipeline found." \
                  "If you want to include a dataset not part of omnibenchmark, please use the respective renku commands." \
                  "CAUTION: External datasets will not be automatically updated by omnibenchmark."
   fi
 
   echo "${success_ids[@]}"
}

