#!/usr/bin/env bash

##############################################################################################
#                          Dataset Utils Functions                                           #
#                                                                                            #
# Functions to facilitate dataset handling related  with renku and the renku knowledge base. #
# Most functions are generalizable, but some specifically target omnibenchmark scripts and   #
# are only valid or useful in an omnibenchmark context/repository                            #
#                                                                                            #
# Change History                                                                             #
# 18/04/2021  Oksana Riba Grognuz    Original code: functions to interact with datasets and  #
#                                    query renku's knowledge base using their user endpoint. #
# 10/05/2021  Almut Lütge            Major update: Add functions to handle datasets and      #
#                                    manage dataset files.                                   #
#                                                                                            #
# authors: Oksana Riba Grognuz & Almut Lütge                                                 #
##############################################################################################
##############################################################################################
##############################################################################################

##############################################################################################

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Prerequisites  ----------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}

{ # try
    source src/utils/bash/check_datasets.sh
} || { 
    echo "Warning, src/utils/check_datasets.sh does not exist and may be required for some functions."
}

# Error function
error_exit()
{
  printf '%s\n' "$@" 1>&2
  exit 1
} 


## --------------------------------------------------------------------------------------- ##
## -------------------------------- Global variables  -------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/global_vars.sh
} || { 
    echo "Warning, src/utils/global_vars.sh does not exist and may be required for some functions."
}



##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##


help_query_datasets_by_string()
{
   echo "Function to query the renku knowledge base endpoint for a specific string. Will return the IDs (default) or defined field of all datasets with this string in their name or keywords."
   echo
   echo "Syntax: query_datasets_by_string STRING FIELD"
   echo
   echo "Arguments:"
   echo "STRING         String to query the knowledge base for"
   echo "FIELD          Field of the dataset entries to return for matched datasets."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_match_datasets_keywords()
{
   echo "Function to compare a specific keyword with the dataset keyword of the input dataset and return any target dataset features, if keywords match."
   echo
   echo "Syntax: match_datasets_keywords DATASET_ID KEYWORD FIELD"
   echo
   echo "Arguments:"
   echo "DATASET_ID     Dataset ID of for the dataset to match the keyword."
   echo "KEYWORD        Word to match the datasets keyword."
   echo "Field          Field to provide the output for if keywords match."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_query_datasets_by_keyword()
{
   echo "Function to query the renku knowledge base endpoint for all datasets with a specific keyword and returns a specified output field for those that match."
   echo
   echo "Syntax: query_datasets_by_keyword KEYWORD FIELD"
   echo
   echo "Arguments:"
   echo "KEYWORD        Word to match the datasets keyword."
   echo "Field          Field to provide the output for if keywords match."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_datasets_properties_by_id()
{
   echo "Function to return a defined dataset property/field by dataset ID."
   echo
   echo "Syntax: datasets_properties_by_id DATASET_ID FIELD"
   echo
   echo "Arguments:"
   echo "DATASET_ID     Dataset ID"
   echo "Field          Field to provide the output for if keywords match."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}



help_get_sameAs_data_id()
{
   echo "Function to return all data ids of identical datasets including the original dataset id."
   echo
   echo "Syntax: get_sameAs_data_id DATASET_ID"
   echo
   echo "Arguments:"
   echo "DATASET_ID     Dataset ID"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}



help_import_datasets_by_keyword()
{
   echo "Function to import all datasets with a specified keyword, if not already present and update those that are already present."
   echo "This function will import datasets from their original repository (see check_import_from_target -h)"
   echo
   echo "Syntax: import_datasets_by_keyword KEYWORD"
   echo
   echo "Arguments:"
   echo "KEYWORD        Keyword to select datasets for import/update by."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo "o              Orchestrator namespace without gitlab prefix, e.g. defaults to omnibenchmark/orchestrator"
   echo "f              Forces the dataset import and ignores checks on the orchestrator CICD."
   echo
}


help_import_datasets_by_string()
{
   echo "Function to import all datasets with a specified string in their name or keyword, if not already present and update those that are already present."
   echo "CAUTION: This function will NOT neccessarily import datasets from their original repository (see check_import_from_target -h)."
   echo "We recommend to use 'import_datasets_by_keyword' to prevent ambiguity."
   echo
   echo "Syntax: import_datasets_by_string STRING"
   echo
   echo "Arguments:"
   echo "STRING         String to select datasets for import/update by."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_create_dataset()
{
   echo "Function to create a renku dataset with the properties defined by the options or a bash array DATA_VARS (see config.sh in omnibenchmark)"
   echo "If a dataset with the defined name already exists, nothing is done."
   echo
   echo "Syntax: create_dataset [OPTIONS]"
   echo
   echo "Options:"
   echo "n       Dataset name - default DATA_VARS['name']."
   echo "t       Dataset title - default DATA_VARS['title']."
   echo "d       Dataset description - default DATA_VARS['description']."
   echo "k       Dataset keyword(s) as list - default TAGLIST."
   echo "c       Check/dryrun - will print the command, but not execute it."
   echo "h       Print this Help."
   echo
}

help_list_project_datasets_by_keyword()
{
   echo "Function to list datasets in the current project that contain a specified keyword."
   echo
   echo "Syntax: list_project_datasets_by_keyword KEYWORD FIELD"
   echo
   echo "Arguments:"
   echo "KEYWORD      Keyword used to import datasets with."
   echo "FIELD        Field to return - default is identifier"
   echo
   echo "Options:"
   echo "h       Print this Help."
   echo
}


help_get_upstream_dataset()
{
   echo "Function to get the dataset name from where the file in the upstream project was generated"
   echo
   echo "Syntax: get_upstream_dataset FILE"
   echo
   echo "Arguments:"
   echo "FILE      File to find the upstream dataset for"
   echo
   echo "Options:"
   echo "h       Print this Help."
   echo
}

##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Helper functions/snippets -------------------------------##
## ---------------------------------------------------------------------------------------- ##

function get_all_datatset_ids () {

    renku dataset ls --format json | jq -r '.[] | .'\"identifier\"
    
}


function dataset_exist () {

    # Input: Dataset_name
    # Output: Dataset name, if dataset exist, else an empty string
    
    local data_name=$1
    renku dataset ls --format json | jq -r '.[] | .name' | grep -E -x "$data_name"   
}

function find_unique () {
    echo "$@" | tr ' ' '\n' | sort | uniq -u
}



##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ----------------------- Functions to query datasets on the KG ---------------------------##
## ---------------------------------------------------------------------------------------- ##

function query_datasets_by_string () {

    # Input1 String to query
    # Input2 Field to retun - default identifier.
    # Output dataset IDs of all datasets with this string in name or keyword.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_query_datasets_by_string
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    #query
    local string=$1
    local field=${2:-identifier}
    curl -s "${KG_URL}/datasets?query=$string" | jq -r '.[] | .'\""$field"\"''
}


function match_datasets_keywords () {
    
    # Input1 Dataset ID
    # Input2 Keyword to match the datasets keyword
    # Input3 Field to return for matches (default identifier)
    # Output Return fields value, if the datasets keyword matches the queried one.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_match_datasets_keywords
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   # match keywords
   local keyword
   local dataset_id=$1
   keyword=$(echo "$2" | awk '{print tolower($0)}')
   local field=${3:-identifier}
   
   curl -s "${KG_URL}/datasets/$dataset_id" \
   | jq -r 'select(.keywords[] | ascii_downcase == ('\""$keyword"\"')) | .'\""$field"\"
   }
     
   
function query_datasets_by_keyword () {
   
   # Input1 Keyword to query.
   # Input2 Dataset field tor return (Default identifier).
   # Output Return fields value of all datasets matching the queried keyword.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_query_datasets_by_keyword
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
   
   # Query
   local keyword=$1
   local field=${2:-identifier}  
   
   for dataset_id in $(query_datasets_by_string "$keyword")
   do 
       match_datasets_keywords "$dataset_id" "$keyword" "$field"
   done
   }
   

function datasets_properties_by_id () {

   # Input  Dataset ID
   # Output Return fields value of the input dataset.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_dataset_properties_by_id
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
   
   # Query
   local dataset_id=$1
   local field=${2:-name}  
   
   curl -s "${KG_URL}/datasets/$dataset_id" | jq -r '.'\""$field"\"''
   }
   


function get_sameAs_data_id () {

   # Input  Dataset ID
   # Output Return dataset ids listed in the sameAs field as well as the input id .
        
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case "${option}" in
            h) 
                help_get_sameAs_data_id
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
   
   # Query
   local id=$1
   local data_url="${RENKU_URL}/datasets/"
   local out_id=("$id")
   
   same_url=$(curl -s "${KG_URL}/datasets/${id}" | jq -r '.'\"sameAs\" | grep "$data_url")
       if [[ -n $same_url ]] 
       then
           same_id="${same_url//$data_url/}"
           out_id+=("$same_id") 
       fi
       
    echo "${out_id[@]}"
   }




function import_datasets_by_keyword () {

   # Input  Keyword to query datasets for
   # Output Imports all datasets with the input keyword that are not already imported and updates those that are already imported.
    
    local OPTIND o_path="omnibenchmark/orchestrator"
    
    # Display help message
    while getopts ":o:hf" option; do
        case $option in
            h) 
                help_import_datasets_by_keyword
                return 1;;
            o)
                o_path=${OPTARG};;
            f)
                force_import='true';;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
   
   # Import/update datasets
   local keyword=$1
   
   all_ids=$(query_datasets_by_keyword "$keyword" identifier)
   data_url="${RENKU_URL}/datasets/"
   in_ids=()
   
   #include "sameAs" IDs
   for id in ${all_ids}
   do
       in_ids+=( $(get_sameAs_data_id "$id") )  
   done
   
   # Test ids to be from the origin project, have a successful cicd and be part of omnibenchmark
   origin_ids=$(check_import_from_target "${in_ids[@]}")
   if [[ force_import ]]
   then 
       success_ids=${origin_ids[@]}
   else 
       omni_ids=$(check_orchestrator -o "$o_path" "${origin_ids[@]}") 
       success_ids=$(check_project_status -o "$o_path" ${omni_ids[@]})
   fi
   
     
   
   for dataset_id in ${success_ids}
   do
      local data_name
      local data_exist
      local data_url
      data_name=$(datasets_properties_by_id "$dataset_id" name)
      data_exist=$(dataset_exist "$data_name")
      data_url=$(datasets_properties_by_id "$dataset_id" url)
      
      if [[ -n "${data_exist}" ]]
      then
         renku dataset update "$data_name"
      else
         renku dataset import -y "$data_url"
      fi
   done
   }



function import_datasets_by_string () {

   # Input  String to query datasets for
   # Output Imports all datasets with the input string in name, description, title or keyword that are not already imported and updates those that are already imported.
   # Warning: Deprecated - does not test for repository and is not testes. Use 'import_dataset_by_keyword' instead.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_import_datasets_by_string
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
   
   
   #Import/Update
   local keyword=$1
   local data_name
   local data_path
   local data_url
   
   for dataset_id in $(query_datasets_by_string "$keyword" identifier)
   do
      
      data_name=$(datasets_properties_by_id "$dataset_id" name)
      data_path="data/${data_name}"
      data_url=$(datasets_properties_by_id "$dataset_id" url)
      
      if [ -d "${data_path}" ]
      then
         renku dataset update "$data_name"
      else
         renku dataset import -y "$data_url"
      fi
   done
   }


   
function dataset_files_by_id () {

   local dataset_id=$1
   curl -s "${KG_URL}/datasets/${dataset_id}" \
                       | jq -r ".hasPart | .[] | .atLocation"
   }




## ---------------------------------------------------------------------------------------- ##
## ---------------------- Functions to handle datasets using renku -------------------------##
## ---------------------------------------------------------------------------------------- ##

function create_dataset () {

    # Input -n Name of the dataset to create
    # Input -t Title of the dataset to create
    # Input -d Description of the dataset to create
    # Input -k Keywords of the dataset to create
    # Output: Creates a renku dataset with the defined properties, if no dataset with this name exist in this project and on the knowledge base.
    
    # Get defaults from config file
    if [[ -v DATA_VARS[@] ]]
    then
        local create_opt=(-t "${DATA_VARS[title]}" -d "${DATA_VARS[description]}" -k "${TAG_LIST}")
        local name=${DATA_VARS[name]}
    fi
    
    #Get Options
    local OPTIND 
    
    while getopts ":n:t:d:k:hc" o; do
        case "${o}" in
            n)
                name="${OPTARG}"
                create_opt=()
                ;;
            t)
                create_opt+=(-t "${OPTARG}")
                ;;
            d)
                create_opt+=(-d "${OPTARG}")
                ;;
            k)
                create_opt+=(-k "${OPTARG}")
                ;;
            h)
                help_create_dataset
                return 1
                ;;
            c)
                echo "CHECK: The following command would be executed:"
                echo "renku dataset create \"${name}\" ${create_opt[@]}"
                return 1
                ;;
            \?) 
                error_exit "Error: Invalid option";;
        esac
    done
    shift $((OPTIND-1))
    
    # Check if dataset exists in this project
    local data_exist
    local string_exists
    data_exist=$(dataset_exist "$name")
    
    if [[ -z $data_exist ]]  
    then
    
        # Check if dataset already exists in the knowledge base
        string_exists=$(curl -s "${KG_URL}/datasets?query=$name" | jq -r '.[] | .name')
        
        if [[ $string_exists =~ $name ]]
        then
            error_exit "ERROR: A dataset with the name $name does already exist on the knowledge base." \
                       "To prevent ambiguity please check and either import the existing dataset instead or choose a different name."
        fi
        
        echo "Creating renku dataset $name"
        renku dataset create "${name}" "${create_opt[@]}"
        
    else
        echo "Found renku dataset $name"
    fi
   }


function add_files_to_dataset_by_name () {
   
    local dataset_name=$1
    local file_names=("${@:2}")
    local file_exists
   
    for file in "${file_names[@]}"
    do
        file_exists=$(renku dataset ls-files "$dataset_name" -I "${file}" | wc -l)
        
        if [ "$file_exists" -gt 2 ]
        then
            # update file using renku command.
            renku dataset update -I "$file" "$dataset_name"
            #renku dataset add -o $dataset_name $file
        else
            echo "Adding $file to the dataset $dataset_name"
            renku dataset add "${dataset_name}" "${file}"
        fi
    done
   }


## ---------------------------------------------------------------------------------------- ##
## ------------------------ Functions to manage datasets(-files) ---------------------------##
## ---------------------------------------------------------------------------------------- ##

function list_data_by_file_name () {

    # input: file name as specified in config.sh
    # output: all data names with a file of input file name 
    
    local file_name=$1
   
    find_unique $(ls "${IN_PATH[$file_name]}" | grep "${IN_PREFIX[$file_name]}") | sed "s/${IN_PREFIX[$file_name]}_//" | sed "s/.${IN_EXT[$file_name]}//"
   }




function list_project_datasets_by_keyword () {
    
    # Input: Keyword
    # Output: List of all datasets in the current project with a specific keyword.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_list_imported_datasets
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    keyword=$1
    field=${2:-identifier}
    
    project_data=$(renku dataset ls --format json | jq -r '.[] | select( .keywords[] | contains( '\""$keyword"\"' ) ) | .'\""$field"\")
        
    if [[ -z $project_data ]]
    then
        error_exit "ERROR: No imported datasets found!"
    fi
    
    echo "${project_data}"
    
    }


function get_upstream_dataset () {
    
    # Input: Keyword
    # Output: List of all datasets in the current project with a specific keyword.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_upstream_dataset
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    count_file=$1
    
    # Rewrite when KG is available for cli1.0.0
    basename -- "$count_file" | sed "s/${IN_PREFIX['count_file']}_//" | sed "s/.${IN_EXT['count_file']}//"
    
    }
