#!/usr/bin/env bash


##############################################################################################
#                              Method Utils Functions                                        #
#                                                                                            #
# Functions to facilitate running of methods scripts as part of an omnibenchmark.            #
# Most functions are specifically target omnibenchmark scripts and are only valid or useful  #
# in an omnibenchmark context/repository.                                                    #
#                                                                                            #
# Change History                                                                             #
# 10/06/2021  Almut Lütge    Original code: functions to run omnibenchmark methods while     #
#                            exploring their parameter space                                 #
# 20/12/2021  Almut Lütge    Add extensions/adaptations for renku-python version 1.0.0       #
#                                                                                            #
# author: Almut Lütge                                                                        #
##############################################################################################
##############################################################################################
##############################################################################################


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------------- Prerequisites  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/bash/dataset_utils.sh
} || { 
    echo "Warning, utils/bash/dataset_utils.sh does not exist and may be required for some functions."
}

{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}

# Error function
error_exit()
{
  printf '%s\n' "$@" 1>&2
  exit 1
} 


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##

help_find_all_param_json()
{
   echo "Function to find and check the json file to define the overall parameter space. Returns the json file's path."
   echo
   echo "Syntax: find_all_param_json PARAM_KEY"
   echo
   echo "Arguments:"
   echo "PARAM_KEY      Name of the parameter dataset with a json file defining the overall parameter space. This dataset needs to be part of the repository."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_get_method_param_combs()
{
   echo "Function to generate a list of all valid parameter combinations for a method."
   echo
   echo "Syntax: get_method_param_combs PARAM_JSON PARAM_AR"
   echo
   echo "Arguments:"
   echo "PARAM_JSON     Json file with path, that defines the overall parameter space for this benchmark."
   echo "PARAM_AR       Name of the parameter array to define all valid parameter (keys) for this method with their default values (default PARAM_AR as defined in src/config.sh)."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_generate_workflow_json()
{
   echo "Function to generate a json file from all methods specific valid parameter combinations to define workflow settings as used as input for renku workflow loop."
   echo
   echo "Syntax: generate_workflow_json PARAM_LIST"
   echo
   echo "Arguments:"
   echo "PARAM_LIST     List with all valid parameter combinations as generated by get_method_param_combs. See get_method_param_combs -h for help."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_match_wflow_names()
{
   echo "Generates an associative array with the renku workflow names as values and the 'original' parameter/output names as keys. This array can be used to match names from parameter combinations with the actual renku workflow names ."
   echo
   echo "Syntax: match_wflow_names OUT_ARR RENKU_WNAME1 RENKU_WNAME2"
   echo
   echo "Arguments:"
   echo "OUT_ARR          Name of the associative array to store matches in."
   echo "RENKU_WNAME1     Renku workflow name of parameter1."
   echo "RENKU_WNAME2.X   Renku workflow name of parameter2..X."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_match_wflow_params()
{
   echo "Generates an associative array with the renku workflow names as keys and the respective parameter values from a specific entry within the parameter.json file."
   echo
   echo "Syntax: match_wflow_params OUT_ARR PARAM_JSON INDEX PARAM_NAME_ARR"
   echo
   echo "Arguments:"
   echo "OUT_ARR          Name of the associative array to output"
   echo "PARAM_JSON       Json file with all valid and not yet executed parameter combinations."
   echo "INDEX            Index from the parameter json to extract parameter values from. Caution - index + 1."
   echo "PARAM_NAME_ARR   Associative arry to match renku workflow names and 'original' names (see help_match_wflow_names)"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_get_out_name_from_parameter()
{
   echo "Generates an associative array with output file categories and corresponding output file names for a specified dataset and parameter combination."
   echo
   echo "Syntax: get_out_name_from_parameter OUT_ARR DATA_NAM PARAM_ARR OUT_PRE OUT_EXT OUT_PATH"
   echo
   echo "Arguments:"
   echo "OUT_ARR          Name of the associative array to output"
   echo "DATA_NAM         Name of the dataset to generate the output file name for."
   echo "PARAM_ARR        Name of the associative array with the parameter values and names."
   echo "OUT_PRE          Name of the associative array, that specifies the prefix of all output files. Usually specified in src/config.sh."
   echo "OUT_EXT          Name of the associative array, that specifies the extensions of all output files. Usually specified in src/config.sh."
   echo "OUT_PATH         Path to write output files to (path to use as dirname)."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_match_wflow_outputs()
{
   echo "Generates an associative array with renku wflow output names and corresponding output file names for a specified dataset and parameter combination."
   echo
   echo "Syntax: get_out_name_from_parameter OUT_ARR DATA_NAM PARAM_ARR OUT_PRE OUT_EXT OUT_PATH PARAM_NAME_ARR"
   echo
   echo "Arguments:"
   echo "OUT_ARR          Name of the associative array to output"
   echo "DATA_NAM         Name of the dataset to generate the output file name for."
   echo "PARAM_ARR        Name of the associative array with the parameter values and names."
   echo "OUT_PRE          Name of the associative array, that specifies the prefix of all output files. Usually specified in src/config.sh."
   echo "OUT_EXT          Name of the associative array, that specifies the extensions of all output files. Usually specified in src/config.sh."
   echo "OUT_PATH         Path to write output files to (path to use as dirname)."
   echo "PARAM_NAME_ARR   Associative arry to match renku workflow names and 'original' names (see help_match_wflow_names)"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_parse_wflow_params()
{
   echo "Generates a list of parsed parameter from an associative array, that can be used as input to 'renku workflow execute'"
   echo
   echo "Syntax: parse_wflow_params PARAM_ARR"
   echo
   echo "Arguments:"
   echo "PARAM_ARR        Name of the associative array with the parameter and coresponding values to parse."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_get_workflow_feature()
{
   echo "Get name of specified features from a specified workflow description"
   echo
   echo "Syntax: get_workflow_feature FEATURE WORKFLOW"
   echo
   echo "Arguments:"
   echo "FEATURE        Name of feature to get all names from"
   echo "WORKFLOW       Name of the workflow to get feature names from"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_get_workflow_value()
{
   echo "Get value of specified features from a specified workflow description"
   echo
   echo "Syntax: get_workflow_feature FEATURE WORKFLOW"
   echo
   echo "Arguments:"
   echo "FEATURE        Name of feature to get all values for"
   echo "WORKFLOW       Name of the workflow to get feature values for"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_get_comb_number()
{
   echo "Function to get the number of parameter combinations defined in a specified parameter json file"
   echo
   echo "Syntax: get_comb_number PARAM_JSON"
   echo
   echo "Arguments:"
   echo "PARAM_JSON     Json file to specify all valid parameter combinations by the key 'valid_combinations'"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_invalid_parameter_json()
{
   echo "Generates a json file with one key called 'invalid_combinations', if JSON_FILE does not exist."
   echo
   echo "Syntax: invalid_parameter_json JSON_FILE"
   echo
   echo "Arguments:"
   echo "JSON_FILE      Path to the json file."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_filter_parameter_json()
{
   echo "Filters a json file (PARAM_JSON) with parameter combinations by combinations specified by FILTER_KEY in another json filer (FILTER_JSON)."
   echo
   echo "Syntax: filter_parameter_json PARAM_JSON FILTER_KEY FILTER_JSON"
   echo
   echo "Arguments:"
   echo "PARAM_JSON     Json file to be filtered."
   echo "FILTER_KEY     Key that specifies the combinations to be filtered."
   echo "FILTER_JSON     Json file that contains the filter combinations, specified by FILTER_KEY."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_filter_complete_wflows()
{
   echo "Function to filter an array of workflows by those that are associated to at least the number of output files that are expected by the number of parameter combinations specified in the Parameter json and not excluded by another json file with combinations to filter."
   echo
   echo "Syntax: filter_complete_wflows PARAM_JSON WFLOW_ARR FILTER_JSON OUT_PATH OUT_PRE"
   echo
   echo "Arguments:"
   echo "PARAM_JSON     Json file containing all parameter combinations valid for all workflows."
   echo "WFLOW_ARR      Array that contains all workflows asociated with the project."
   echo "FILTER_JSON    Json file specifying parameter combinations to exclude for specific workflows."
   echo "OUT_PATH       Path to the workflow output files"
   echo "OUT_PRE        Array to output file prefixes, to associuate output files to single workflow executions."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_check_all_files_existing()
{
   echo "Takes a list of files and returns a logical indicating if ALL of the files exist. Will return 'false', if any input file is missing. "
   echo
   echo "Syntax: check_all_files_existing FILE_1 .."
   echo
   echo "Arguments:"
   echo "FILE_1 ..      Files to check for existance."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_filter_parameter_thresholds()
{
   echo "Filters a parameter json file, with parameter names as keys and values as a string separarted by spaces by thresholds defined in a minimum, maximum and deletion assay."
   echo
   echo "Syntax: filter_parameter_thresholds PARAM_ALL_JSON MIN_ARR MAX_ARR DEL_ARR"
   echo
   echo "Arguments:"
   echo "PARAM_ALL_JSON   Json file with parameter names as key and values combined in one string."
   echo "MIN_ARR          Array with the parameter names as key and the minimum thresholds as value."
   echo "MAX_ARR          Array with the parameter names as key and the maximum thresholds as value."
   echo "DEL_ARR          Array with the parameter names as key and the all values to exclude as values."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}

##############################################################################################

function dataset_exist () {

    # Input: Dataset_name
    # Output: Dataset name, if dataset exist, else an empty string
    
    local data_name=$1
    renku dataset ls --format json | jq -r '.[] | .name' | grep -E -x "$data_name"   
}


## ---------------------------------------------------------------------------------------- ##
## --------------------- Functions to access parameter space  ------------------------------##
## ---------------------------------------------------------------------------------------- ##


# ---------------------------- find_all_param_json ----------------------------------------- #

function find_all_param_json (){

    # Input1: PARAM_KEY the keyword of the dataset that describes the overall parameter space
    # Output: Json file that defines the overall parameter space. Must be imported into the project.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_find_all_param_json
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define variables and generate output
    local param_key=$1
    local param_name
    param_name=$(query_datasets_by_keyword "$param_key" name | sort | uniq)
    
    if [[ -z $(dataset_exist "$param_name") ]]
    then
        error_exit "Dataset $param_name not found." \
                   "Please check PARAM_KEY in config.sh and/or import $param_name."
    fi
    
    local param_json
    param_json=$(renku dataset ls-files "${param_name}" -c path | grep json)
       
    # Check output is only one file
    local file_length
    file_length=$(echo "${param_json}" | wc -w)
    
    case $file_length in
            0) 
                error_exit "ERROR: Parameter file does not exist. Please check parameter dataset." \
                            "Detected: ${param_json}";;
            1)
                echo "${param_json}";;
            
            \?) 
                error_exit "ERROR: Multiple json files are found as part of the parameter dataset." \
                           "Make sure the dataset contains only one json file." \
                           "The following files were detected:" \
                           "${param_json}";;
            esac
}


# ------------------------------- get_all_param_combs --------------------------------- #


function get_method_param_combs (){
    
    # Input1: Path to the parameter.json file that defines the overall parameter space for a benchmark.
    # Input2: Name of the parameter array, that defines the valid parameter for a specific method.
    # Output: List with all valid parameter combinations
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_method_param_combs
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define variables
    local param_json=$1
    local -n param_ar=${2:-PARAM_AR}
    
    # Filter parameter names
    json_keys=$(cat "$param_json" | jq -r 'keys | .[]')
    json_filtered=$(cat "${param_json}" | jq -r '{'$(echo ${!param_ar[@]} | tr " " ",")'}')
    
    # Generate json array with all parameter combinations
    json_ar=$(echo ${json_filtered} | jq -r 'map_values(.[] | split(" ")) | [.[]] | map(select(length > 0)) | combinations | to_entries')
    keys_sorted=$(echo ${json_filtered} | jq -r 'map_values( .[] | select(length > 0)) | keys_unsorted')
    
    # Replace array keys by parameter names
    for param in "${!param_ar[@]}"; do
        
        # Check if parameter is defined in parmeter json
        if [[ ! "${json_keys[@]}" =~ (^|[[:space:]])$param($|[[:space:]]) ]];
        then
            error_exit "Could not find values for parameter $param." \
                       "Please make sure $param and all possible values are specified in $param_json." \
                       "As $param_json was imported this should be done in this benchmarks parameter repo."
        fi
       
        param_index=$(echo $keys_sorted | jq -r 'map(. == "'$param'") | index(true)')
        json_ar=$(echo $json_ar | jq -r 'map( if .key == '$param_index' then .key |= "'$param'" else . end )')
    
    done

    json_obj=$(echo $json_ar | jq -r 'from_entries')
    json_list=$(echo $json_obj | jq -s 'reduce inputs as $in (.; . + $in)')
    
    echo $json_list | jq -r '{"valid_combinations": .}'
}


# --------------------------- generate_workflow_json ---------------------------------- #

function generate_workflow_json (){
    # input parameter combinations as from get_method_param_combs
    # output workflow description json to use with renku workflow loop
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_parse_default_params_to_Rargs
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define inputs
    local param_all="$*"
    local json_key="valid_combinations"
    
    # Generate output
    key_val=$(echo {\""$param_all"\"} | sed "s/ /\" }, { \"/g" | sed "s/=/\": \"/g" | sed "s/_/\", \"/g") 
    
    echo -e "{ \"$json_key\": [$key_val]}\n"   
}



# --------------------------- match_wflow_names ---------------------------------- #

function match_wflow_names (){
    # input: renku workflow names of parameter and outputs
    # associative array with matching workflow names and variable names
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_match_wflow_names
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define inputs
    local -n match=$1
    local w_names=${@:2}
    
    # Generate output
    for w_nam in ${w_names[@]}; do
        nam_striped="${w_nam//-[0-9]*/}"
        match["$nam_striped"]="$w_nam"
    done
    
}


# --------------------------- match_wflow_params ---------------------------------- #

function match_wflow_params (){
    # input: parameter json, index and wflow name array
    # associative array with matching workflow names and parameter values
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_match_wflow_params
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define inputs
    local -n match=$1
    local param_fi=$2
    local index=$3
    local -n names=${4:-param_names}
    
    # Generate output
    keys=$(cat $param_fi | jq -r '.valid_combinations | .['$index-1'] | keys | .[]')
    
    for key in ${keys[@]}; do
        val=$(cat $param_fi | jq -r '.valid_combinations | .['$index-1'] | .'\"$key\")
        w_nam="${param_names["$key"]}"
        match["$w_nam"]="$val"
    done
    
}


# --------------------- get_out_name_from_parameter --------------------------- #

function get_out_name_from_parameter(){

    # Input: Output array name, dataset name, parameter array, array with prefix and extension of outfiles and outpath 
    # Output: Assoicative array with R arguments-values of inputs and parameter defaults.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_out_name_from_parameter
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    local -n out_names=$1
    local data=$2
    local -n arr=${3:-PARAM_AR}
    local -n out_pre=${4:-OUT_PREFIX}
    local -n out_ext=${5:-OUT_EXT}
    local out_path=${6:-$OUT_PATH}
    local all_param=()
    local all_values=()
    
    for param in "${!arr[@]}"; do
        all_param+=("$param")
    done
    
    IFS=$'\n' sorted=($(sort <<<"${all_param[*]}")); unset IFS
    for par in "${sorted[@]}"; do all_values+=("${arr["$par"]}"); done
    
    param_parsed=$(printf "%s_" "${all_values[@]}")
    
    for out_file in "${!out_pre[@]}"; do
        pre="${out_pre["$out_file"]}"
        ext="${out_ext["$out_file"]}"
        out_names["$out_file"]="${out_path}/${data}_${param_parsed}${pre}${ext}"
    done
    
}


# --------------------- match_wflow_outputs --------------------------- #

function match_wflow_outputs (){
    # input: renku workflow names outputs
    # associative array with matching workflow names and variable names
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) help_match_wflow_outputs
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define inputs
    local -n match=$1
    local data=$2
    local -n arr_n=${3:-wflow_params}
    local -n out_prefix=${4:-OUT_PREFIX}
    local -n out_exten=${5:-OUT_EXT}
    local out_path=${6:-$OUT_PATH}
    local -n w_nam_arr=${7:-param_names}
    
    declare -A var_nam_arr
    
    # Get output names
    get_out_name_from_parameter var_nam_arr "$data" arr_n out_prefix out_exten "$out_path"
    
    # Match with renku workflow names
    for var_nam in ${!var_nam_arr[@]}; do
        w_nam="${w_nam_arr["$var_nam"]}"
        match["$w_nam"]="${var_nam_arr["$var_nam"]}"
    done
    
}


# --------------------- parse_wflow_params --------------------------- #

function parse_wflow_params(){

    # Input: Name of assoicative array with R arguments-values
    # Output: Prints renku input from array with R arguments and input array keys defined in config.sh
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_parse_wflow_params
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local -n arr=$1
    
    for par in ${!arr[@]}
    do
       echo "--set ${par}=${arr["$par"]}"
    done | tr '\n' ' '
}


# --------------------- get_workflow_feature --------------------------- #

function get_workflow_feature(){

    # Input1: Feature to extract from workflow description
    # Input2: Workflow name
    # Output: Extracts all feature names from a specified feature of a specified workflow
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_workflow_feature
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local feature=$1
    local workflow=$2
    
    res=$(renku workflow ls --format json-ld | jq -r '.[] | select(.'\"http://schema.org/name\"' | contains('\"$workflow\"')) | .'\"https://swissdatasciencecenter.github.io/renku-ontology#$feature\"' | .[] | .'\"http://schema.org/name\")
    
    printf '%s\n' ${res[@]}
}


# --------------------- get_workflow_value --------------------------- #

function get_workflow_value(){

    # Input1: Feature to extract from workflow description
    # Input2: Workflow name
    # Output: Extracts default values from a specified feature of a specified workflow
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_workflow_feature
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local feature=$1
    local workflow=$2
    
    res=$(renku workflow ls --format json-ld | jq -r '.[] | select(.'\"http://schema.org/name\"' | contains('\"$workflow\"')) | .'\"https://swissdatasciencecenter.github.io/renku-ontology#$feature\"' | .[] | .'\"http://schema.org/defaultValue\")
    
    printf '%s\n' ${res[@]}
}


# --------------------- get_comb_number --------------------------- #

function get_comb_number(){

    # Input1: Filtered parameter json file
    # Output: Number of missing parameter combinations to run 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_comb_number
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local filtered_json=$1
    
    cat $filtered_json | jq -r '.valid_combinations | length'
}


## ---------------------------------------------------------------------------------------- ##
## --------------------- Functions to filter parameter space  ------------------------------##
## ---------------------------------------------------------------------------------------- ##


# --------------------- invalid_parameter_json --------------------------- #

function invalid_parameter_json(){

    # Input: Path to the json file to contain invalid parameter combinations
    # Output: Generates a json file to add invalid parameter combinations to 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_invalid_parameter_json
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local json_path=$1
    
    if [[ ! -f $json_path ]]
    then
        echo -e "{ \"invalid_combinations\": []}\n" > $json_path 
    fi
}


# --------------------- filter_parameter_json --------------------------- #

function filter_parameter_json(){

    # Input1: Path to the parameter json file to contain all parameter combinations
    # Input2: Path to the parameter json file to contain parameter combinations to be filtered
    # Input3: Key to access the combinations to filter from within Input2 
    # Output: Generates a json file with all parameter combinations from Input2 with Input3 as key filtered.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_filter_parameter_json
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local param_json=$1
    local filter_key=${2:-invalid_combinations}
    local filter_json=${3:-$PARAM_EXCL}
    
    
    ## Filter param_json by filter_json
    excl_ind=$(cat ${param_json} | jq --slurpfile invalid ${filter_json} '.valid_combinations | to_entries | map(select(.value | contains($invalid | .[].'$filter_key' | to_entries? | .[].value))) | .[].key')
    
    if [[ -z $excl_ind ]]
    then 
        cat ${param_json} | jq -r
    else
        cat ${param_json} | jq -r 'del(.valid_combinations | .['$(echo $excl_ind | tr " " ",")'])'
    fi
}


# --------------------- filter_complete_wflows --------------------------- #

function filter_complete_wflows(){

    # Input1: Path to the parameter json file to contain all parameter combinations
    # Input2: Arrays with all workflows with a project
    # Input3: Path to the parameter json file to contain parameter combinations to be filtered 
    # Input4: Path to store output files 
    # Output: Returns all wflows with less result files than valid parameter combinations are available.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_filter_complete_wflows
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local param_json=$1
    local -n all_wflows=$2
    local filter_json=${3:-$PARAM_EXCL}
    local out_path=${4:-$OUT_PATH}
    local -n out_pre=${5:-OUT_PREFIX}
    
    
    wflow_rep_all=$(cat $param_json | jq -r '.valid_combinations | length')
    wflow_filtered=()
    
    #filter workflow for complete ones
    for wflow in ${all_wflows[@]}
    do
        wflow_out_all=$(ls $out_path | grep $wflow | wc -l) ## rewrite as soon as KG is available
        wflow_rounds=$(($wflow_out_all/${#out_pre[@]}))
        wflow_excl_spec=$(cat $filter_json | jq -r '.'$wflow'? | length')
        wflow_rep=$((${wflow_rep_all}-${wflow_excl_spec}))
        
        #Check if there are less output files than valid parameter combinations
        if [[ $wflow_rounds -lt $wflow_rep ]]
        then
            wflow_filtered+=("$wflow")
        fi
    done   
    
    echo ${wflow_filtered[@]}
}


# --------------------- check_files_existing --------------------------- #

function check_all_files_existing(){

    # Input: Files to test for existance
    # Output: Logical, if all files exist.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_check_all_files_existing
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local test_files=$@
    local exist=true
        
    for file in ${test_files[@]}
    do
        if [[ ! -f $file ]]
        then
            exist=false
        fi
    done
    
    echo $exist
}



# --------------------- filter_parameter_thresholds --------------------------- #

function filter_parameter_thresholds(){

    # Input: Path to the json file to contain invalid parameter combinations
    # Output: Generates a json file to add invalid parameter combinations to 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_filter_parameter_thresholds
                return 1;;
            \?) 
                echo "Error: Invalid option"
                return 1;;
            esac
    done
    shift $((OPTIND-1))
    
    #parse inputs
    local param_all_json=$1
    local -n min_arr=${2:-PARAM_MIN}
    local -n max_arr=${3:-PARAM_MAX}
    local -n del_arr=${4:-PARAM_DEL}
    
    res_json=$(cat $param_all_json)
    
    # Filter minimums
    for min_key in ${!min_arr[@]}
    do
         filtered_min=$(echo $res_json | jq -r '.'$min_key' | .[] | split(" ") | .[] | tonumber | select(. >='${min_arr["$min_key"]}') | tostring')
         quoted_min=$(echo \"$filtered_min\")
         res_json=$(echo $res_json | jq -r --argjson val "$quoted_min" '.'$min_key'=[$val]')
    done
    
    # Filter maximums
    for max_key in ${!max_arr[@]}
    do
         filtered_max=$(echo $res_json | jq -r '.'$max_key' | .[] | split(" ") | .[] | tonumber | select(. <='${max_arr["$max_key"]}') | tostring')
         quoted_max=$(echo \"$filtered_max\")
         res_json=$(echo $res_json | jq -r --argjson val "$quoted_max" '.'$max_key'=[$val]')
    done
    
    # Filter values
    for del_key in ${!del_arr[@]}
    do
         values=$(echo $(printf "\"%s\" " ${del_arr["$del_key"]}) | tr " " ",")
         filtered_del=$(echo $res_json | jq -r '.'$del_key' | .[] | split(" ") | .[] | del(select(. == ('$values'))) | select(. != null)')
         quoted_del=$(echo \"$filtered_del\")
         res_json=$(echo $res_json | jq -r --argjson val "$quoted_del" '.'$del_key'=[$val]')
    done
    
    echo $res_json | jq -r
}



##############################################################################################
