#!/usr/bin/env bash


##############################################################################################
#                              Metric Utils Functions                                        #
#                                                                                            #
# Functions to facilitate running of methods scripts as part of an omnibenchmark.            #
# Most functions are specifically target omnibenchmark scripts and are only valid or useful  #
# in an omnibenchmark context/repository.                                                    #
#                                                                                            #
# Change History                                                                             #
# 23/06/2021  Almut Lütge    Original code: functions to run omnibenchmark metrics on        #
#                            method results.                                                 #
#                                                                                            #
# author: Almut Lütge                                                                        #
##############################################################################################
##############################################################################################
##############################################################################################


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------------- Prerequisites  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##


{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}

# Error function
error_exit()
{
  printf '%s\n' "$@" 1>&2
  exit 1
} 


## --------------------------------------------------------------------------------------- ##
## -------------------------------- Global variables  -------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/global_vars.sh
} || { 
    echo "Warning, src/utils/global_vars.sh does not exist and may be required for some functions."
}



##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##

help_get_dataset_from_method_result()
{
   echo "Function to extract the dataset name from the name of a method result by matching it with a list of all dataset names."
   echo
   echo "Syntax: get_dataset_from_method_result METH_RES_NAME ALL_DATA"
   echo
   echo "Arguments:"
   echo "METH_RES_NAME      Name of the method output/result file."
   echo "ALL_DATA           List of all possible dataset names (e.g. from meta files)"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_get_run_from_method_result()
{
   echo "Function to extract the method run details about dataset and parameter settings from the metric input file"
   echo
   echo "Syntax: get_run_from_method_result METRIC_INPUT"
   echo
   echo "Arguments:"
   echo "METRIC_INPUT      Name of the method output file that isthe input for metric calculations."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_transform_to_renku_name()
{
   echo "Function to transform a string into a string that is compatible with renkus name and parameter regulations."
   echo "This means a string with lower case letters only and without double '__'"
   echo
   echo "Syntax: transform_to_renku_name INPUT_STRING"
   echo
   echo "Arguments:"
   echo "INPUT_STRING   String to be converted."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_find_metric_info_files()
{
   echo "Function to find metric info files within the input datasets by querying the KB for files generated by METRIC_GEN_SCRIPT."
   echo "METRIC_GEN_SCRIPT should either be defined in 'src/config.sh' or will default to 'src/workflow/generate_metric_info.sh'"
   echo
   echo "Syntax: find_metric_info_files DATASET_IDS"
   echo
   echo "Arguments:"
   echo "DATASET_IDS   IDs of datasets that have contain results from an omni metric run."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_find_metric_result_files()
{
   echo "Function to find metric result files within the input datasets by querying by keyword imported datasets and exclude metric info files."
   echo
   echo "Syntax: find_metric_result_files KEYWORD METRIC_INFO_AR"
   echo
   echo "Arguments:"
   echo "KEYWORD          Keyword used to import metric result datasets."
   echo "METRIC_INFO_AR   Name of the array containing metric info files." 
   echo
   echo "Options:"
   echo "h                Print this Help."
   echo
}

help_get_metafile_from_method_result()
{
   echo "Function to extract the metadata file from the name of a method result by checking corresponding edges."
   echo
   echo "Syntax: get_metafile_from_method_result METH_RES_NAME ALL_META_FILES"
   echo
   echo "Arguments:"
   echo "METH_RES_NAME      Name of the method output/result file."
   echo "ALL_META_FILES     List of all possible meta data files in the current project"
   echo
   echo "Options:"
   echo "h                  Print this Help."
   echo
}

help_get_input_keyword_from_file()
{
   echo "Function to extract the keyword associated to the dataset that the input file is part from."
   echo
   echo "Syntax: get_input_keyword_from_file INPUT_FILE"
   echo
   echo "Arguments:"
   echo "INPUT_FILE      Name of the file to extract the keyword for."
   echo
   echo "Options:"
   echo "h               Print this Help."
   echo
}


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ----------------- Functions to format/work with method results  -------------------------##
## ---------------------------------------------------------------------------------------- ##

# ------------------------ get_dataset_from_method_result ---------------------------------- #

function get_dataset_from_method_result (){

    # Input1: METH_RES_NAME name of the method result file.
    # Input1: ALL_DATA list of all possible dataset names.
    # Output: String with the dataset name of the input method result file.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_dataset_from_method_result
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define variables and generate output
    local meth_res_nam=$1
    local meth_res_name
    meth_res_name=$(basename -- "$meth_res_nam")
    local all_data=${@:2}
    pattern_data=$(echo "${all_data[@]}" | tr ' ' '|')
    dataset_name=$(printf '%s\n' "${meth_res_name}" | grep -E -o "$pattern_data")
    
    # Check only one dataset name has been found
    local name_length
    name_length=$(echo "${dataset_name}" | wc -w)
    
    case $name_length in
            0) 
                error_exit "ERROR: Meta data file for the dataset of this method result not found." \
                            "Name of the method result: ${meth_res_name}";;
            1)
                echo "${dataset_name}";;
            
            \?) 
                error_exit "ERROR: Multiple dataset names have been associated with this method output file." \
                           "Make sure you have non-ambigous names." \
                           "The following names were detected:" \
                           "${dataset_name}";;
            esac
}



# ------------------------ get_run_from_method_result ---------------------------------- #

function get_run_from_method_result () {

    # Input: Method result file that is the metric input.
    # Output: String with method run details with parameter settings and dataset name 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_run_from_method_result
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Get method run information string
    local metric_input_file=$1
   
    basename "$metric_input_file" ".${IN_EXT["${METRIC_INPUT}"]}" | sed 's/_'${IN_PREFIX["${METRIC_INPUT}"]}'//'
   }



function transform_to_renku_name () {

    # Input: String that should be transformed
    # Output: Renku compatible string, with double _ removed and all set to lower case 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_transform_to_renku_name
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # transform
    local input_string
    input_string=$(echo "$1" | awk '{print tolower($0)}')
    echo "${input_string//__/_}"
    
   }



function find_metric_info_files () {

    # Input: Dataset IDs from metric output datasets
    # Output: Metric info files
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_find_metric_info_files
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   local dataset_ids=$*
   local file="${METRIC_GEN_SCRIPT:-src/workflow/generate_metric_info.sh}"
   declare -a metric_info_files
     
   # select data creation repository as source
   for id in ${dataset_ids}
   do
              
       # find metric output project to perform the query in
       data_url="${RENKU_URL}/datasets/"
       origin_id=$(curl -s "${KG_URL}/datasets/$id" | jq -r '.'\"sameAs\" | sed 's,'"${data_url}"',,')
       
       if [[ "${origin_id}" =~ null ]]
       then
           origin_id=$id
       fi
       
       project_path=$(curl -s "${KG_URL}/datasets/$origin_id" | \
           jq -r '.'\"project\"' | .'\"path\")
       
       # find file that was generated by "$file"
       info_file=$(curl -X POST -v -H "Content-Type: application/json" "${GRAPHQL_URL}" -d \
                   '{ "query": "{ lineage(projectPath: \"'"${project_path}"'\", filePath: \"'"${file}"'\") { nodes { id type } } }"}' \
                   | jq -r '.data | .lineage | .nodes[]? | select(.type | contains('\"File\"')) | select(.id != '\""${file}"\"') | .id')

       metric_info_files+=("$info_file")
       
   done
   
   if [[ -z ${metric_info_files[@]} ]]
   then
        error_exit "ERROR: No metric info file was found. Please check the datasets: ${dataset_ids} for a metric info file." \
                   "If the dataset contains a metric info file, please make sure it was generated by METRIC_GEN_SCRIPT as defined in src/config.sh"
   fi
   
   echo "${metric_info_files[@]}"
   }
   
   
   
   
function find_metric_result_files () {

    # Input1: Keyword of metric dataset
    # Input2: Array name with metric info file names
    # Output: Metric result files 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_find_metric_result_files
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   keyword=$1
   local -n metric_ar=$2
   
   res_data=$(list_project_datasets_by_keyword "$keyword" name)

   res_files=$(list_dataset_files | grep "$(printf '%s\n' ${res_data[@]})" | \
            grep -v "$(printf '%s\n' ${metric_ar[@]})")
   
      
   if [[ -z ${res_files[@]} ]]
   then
        error_exit "ERROR: No metric result files were found. Please check the datasets: ${res_data}."
   fi
   
   echo "${res_files[@]}"
}



function get_metafile_from_method_result (){

    # Input1: METH_RES_NAME name of the method result file.
    # Input1: ALL_META_FILES list of all possible metadata files.
    # Output: String with the metadata file that was involved in generating the result file.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_metafile_from_method_result
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Define variables and generate output
    local meth_res_name=$1
    local meta_files=${@:2}
    
    project_path=$(renku dataset ls-files -I "$meth_res_name" --format json-ld | jq -r '.[] | .'\"http://schema.org/isBasedOn\"' | .'\"http://schema.org/url\" | \
                   sed 's,'"${GIT_URL}/"',,' | sed 's,\.git,,')
    
    if [[ "${project_path}" =~ null ]]
    then
        link_url="${RENKU_URL}/projects/"
        project_path=$(renku dataset ls-files -I "$meth_res_name" --format json-ld | jq -r '.[] | .'\"http://schema.org/isPartOf\"' | .'\"@id\" | \
                       sed 's,'"${link_url}"',,')
    fi
    
    
    lin_files=$(curl -X POST -v -H "Content-Type: application/json" "${GRAPHQL_URL}" -d \
                   '{ "query": "{ lineage(projectPath: \"'"${project_path}"'\", filePath: \"'"${meth_res_name}"'\") { nodes { id type } } }"}' \
                   | jq -r '.data | .lineage | .nodes[]? | select(.type | contains('\"File\"')) | select(.id != '\""${meth_res_name}"\"') | .id')
    
    lin_base=$(basename -a $(printf '"%s\n' "${lin_files}"))
    lin_path=$(dirname "${meth_res_name}")
    lin_pattern=$(echo "${lin_base[@]}" | tr ' ' '|')
        
    meta_file=$(printf '%s\n' "$meta_files" | grep -E "$lin_pattern" | grep -E "$lin_path")
        
    # Check only one dataset name has been found
    local name_length
    name_length=$(echo "${meta_file}" | wc -w)
    
    case $name_length in
            0) 
                error_exit "ERROR: Meta data file for the dataset of this method result not found." \
                            "Name of the method result: ${meth_res_name}";;
            1)
                echo "${meta_file}";;
            
            \?) 
                error_exit "ERROR: Multiple meta data files have been associated with this method output file." \
                           "Make sure you have non-ambigous names." \
                           "The following files were detected:" \
                           "${meta_file}";;
            esac
}



function get_input_keyword_from_file () {

    # Input: String that should be transformed
    # Output: Renku compatible string, with double _ removed and all set to lower case 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_input_keyword_from_file
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # transform
    local input_file=$1
    
    dataset=$(renku dataset ls-files -I "$input_file" --format json | jq -r '.[] | .'\"dataset_name\")
    keyword=$(renku dataset ls --format json | jq -r '.[] | select(.['\"name\"'] == '\""$dataset"\"') | .'\"keywords\"'[]')
    
    if [[ -z ${keyword} ]]
    then
        error_exit "No keywords found associated to ${input_file} from dataset ${dataset}."
    fi
   
    echo "${keyword[@]}"
    
   }

