#!/usr/bin/env bash

##############################################################################################
#                          Overview Utils Functions                                          #
#                                                                                            #
# Functions to facilitate generating overview summarize of the renku knowledge base.         #
# Most functions are generalizable, but some specifically target omnibenchmark scripts and   #
# are only valid or useful in an omnibenchmark context/repository                            #
#                                                                                            #
# Change History                                                                             #
# 10/09/2021  Almut Lütge    Original code: functions to explore and query the renku         #
#                                    knowledge base using their user endpoint.               #           
#                                                                                            #
# authors: Almut Lütge                                                                       #
##############################################################################################
##############################################################################################
##############################################################################################

##############################################################################################

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Prerequisites  ----------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}


# Error function
error_exit()
{
  printf '%s\n' "$@" 1>&2
  exit 1
} 


## --------------------------------------------------------------------------------------- ##
## -------------------------------- Global variables  -------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/global_vars.sh
} || { 
    echo "Warning, src/utils/global_vars.sh does not exist and may be required for some functions."
}



##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##


help_get_project_datasets_by_path()
{
   echo "Function to get the FIELD information of all datasets that are part of a project."
   echo
   echo "Syntax: get_project_datasets_by_path PROJECT_PATH FIELD"
   echo
   echo "Arguments:"
   echo "PROJECT_PATH   Path (without gitlab path) to the project to query."
   echo "FIELD          Field of the dataset entries to return - default name."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_get_json_from_arr()
{
   echo "Function to transform bash array into json format."
   echo
   echo "Syntax: get_json_from_arr ARR"
   echo
   echo "Arguments:"
   echo "ARR      Name of the array to convert"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Project related queries ---------------------------------##
## ---------------------------------------------------------------------------------------- ##


function get_project_datasets_by_path () {

    # Input1 String to query
    # Input2 Field to retun - default identifier.
    # Output dataset IDs of all datasets with this string in name or keyword.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_project_datasets_by_path
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    #query
    local project_path=$1
    local field=${2:-name}
    curl -s "${KG_URL}/projects/$project_path/datasets" | jq -r '.[] | .'"$field" | uniq
}




function get_json_from_arr () {

    # Input1 String to query
    # Input2 Field to retun - default identifier.
    # Output dataset IDs of all datasets with this string in name or keyword.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_json_from_arr
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    #query
    local -n arr=$1
    local json_out='{}'
   
   for key_a in ${!arr[@]}
   do
       json_out=$(jq -n --arg key "$key_a" --argjson out "$json_out" '$out + {($key): []}')
       
       for val in ${arr["${key_a}"]}
       do
           json_out=$(echo "$json_out" | jq --arg key "$key_a" --arg val "$val" '.[$key] += [$val]')
       done
   done
   
   echo -e "${json_out}\n"
}
