#!/usr/bin/env bash

##############################################################################################
#                          Provenance Utils Functions                                        #
#                                                                                            #
# Functions to facilitate provenance and related Knowledge graph related actions with renku. #
# Most functions are generalizable, but some specifically target omnibenchmark scripts and   #
# are only valid or useful in an omnibenchmark context/repository                            #
#                                                                                            #
# Change History                                                                             #
# 20/04/2021  Oksana Riba Grognuz    Original code: functions to interact with provenance    #
#                                    graph and query renku's knowledge base                  #
# 20/05/2021  Almut Lütge            Major update: Update functions to work with new renku   #
#                                    onthology in renku 0.15.1                               #
# 10/06/2021  Almut Lütge            Major update: Update Documentation and add helper       #
#                                    functions.                                              #
#                                                                                            #
#                                                                                            #
# authors: Oksana Riba Grognuz & Almut Lütge                                                 #
##############################################################################################
##############################################################################################
##############################################################################################

##############################################################################################

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Prerequisites  ----------------------------------------##
## --------------------------------------------------------------------------------------- ##


{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}


# Error function
error_exit()
{
  printf '%s\n' "$@" 1>&2
  exit 1
}

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Global variables  -------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/global_vars.sh
} || { 
    echo "Warning, src/utils/global_vars.sh does not exist and may be required for some functions."
}


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##

help_list_dataset_files()
{
   echo "Function to list all dataset files in the current repository, independent of the dataset."
   echo
   echo "Syntax: help_list_dataset_files"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_list_plan_inputs()
{
   echo "Function to list all existing workflow inputs containing a specific PATTERN."
   echo
   echo "Syntax: list_plan_inputs PATTERN"
   echo
   echo "Arguments:"
   echo "PATTERN        Pattern to filter plan inputs for. Default is data."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_list_plan_outputs()
{
   echo "Function to list all existing workflow outputs containing a specific PATTERN."
   echo
   echo "Syntax: list_plan_outputs PATTERN"
   echo
   echo "Arguments:"
   echo "PATTERN        Pattern to filter plan outputs for. Default is data."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_find_files_to_process()
{
   echo "Function to list files that are defined as input in src/config.sh and part of a dataset that is associated to the current project,"
   echo "but not yet input of a workflow."
   echo
   echo "Syntax: find_files_to_process PATTERN PATTERN_EXCL"
   echo
   echo "Arguments:"
   echo "PATTERN        '|' separated list of input file name pattern to process in workflows."
   echo "PATTERN_EXCL   '|' separated list of file name pattern not to process in workflows."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_find_datasets_to_process()
{
   echo "Function to list datasets associated to the current project with an input defined in src/config.sh, but not yet input of a workflow."
   echo
   echo "Syntax: find_datasets_to_process"
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_run_command_on_unprocessed()
{
   echo "Function to run a command on all unprocessed files. The command needs to be valid for all unprocessed files. See find_files_to_process -h."
   echo
   echo "Syntax: run_command_on_unprocessed COMMAND PROV_JSON"
   echo
   echo "Arguments:"
   echo "COMMAND        Command to run on unprocessed files. Needs to be valid for all unprocessed input files."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}

help_get_input_file_path()
{
   echo "Function to get the path of the file with specified input index and dataset."
   echo
   echo "Syntax: get_input_file_pat DATASET_NAME INPUT_INDEX"
   echo
   echo "Arguments:"
   echo "DATASET_NAME   Name of the dataset that the file of interest belongs to."
   echo "INPUT_INDEX    Index name of the file of interest. Needs to be a valid key of IN_PREFIX (usually specified in src/config.sh)."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


help_generate_input_json()
{
   echo "Function to write a list of file names to a json file, that can be used as input to run renku."
   echo
   echo "Syntax: generate_input_json FILE_NAMES"
   echo
   echo "Arguments:"
   echo "FILE_NAMES   List of file names to add to the out file."
   echo
   echo "Options:"
   echo "h              Print this Help."
   echo
}


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Helper functions/snippets -------------------------------##
## ---------------------------------------------------------------------------------------- ##

function find_unique () {
    echo "$@" | tr ' ' '\n' | sort | uniq -u
}


function get_project_graph () {
    renku graph export --full --format jsonld > .renku/provenance.json
    renku save
}


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ---------------------- Functions to query KG using access points ------------------------##
## ---------------------------------------------------------------------------------------- ##

function list_dataset_files () {
    
    # Output: List of all datasets files in the current project.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_list_dataset_files
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # List all dataset files
    renku dataset ls-files --format json | jq -r '.[] | .path' | sort | uniq
    
    }


function get_input_file_path(){

    # Input1: Dataset name of the file to get the path
    # Input2: Index name of the file to get the path 
    # Output: Path of the input file with the specified index and dataset.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_input_file_path
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    local data_name=$1
    local input_index=$2
    echo data/"${data_name}"/"${IN_PREFIX[$input_index]}"_*."${IN_EXT[$input_index]}"
}

## ---------------------------------------------------------------------------------------- ##
## -------------------- Functions to query KG using provenance.json ------------------------##
## ---------------------------------------------------------------------------------------- ##


function list_plan_inputs () {

    # Input1: Pattern to filter plan inputs for. Default "data". 
    # Output: All files of pattern $2(default data) listed as inputs for any existing workflow
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_list_plan_inputs
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # find pland inputs
    local pattern=${1:-data}
    renku workflow ls --format json-ld | jq -r '.[] | .'\"https://swissdatasciencecenter.github.io/renku-ontology#hasInputs\"' | .[] | .'\"http://schema.org/defaultValue\" | grep -E "$pattern"
}


function list_plan_outputs () {
    
    # Input1: Pattern to filter plan outputs for. Default "data". 
    # Output: All files of pattern $2(default data) listed as outputs for any existing workflow
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_list_plan_outputs
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    #find plan outputs
    local pattern=${1:-data}
    renku workflow ls --format json-ld | jq -r '.[] | .'\"https://swissdatasciencecenter.github.io/renku-ontology#hasOutputs\"' | .[] | .'\"http://schema.org/defaultValue\" | grep -E "$pattern"
}



function find_files_to_process(){
    
    # Input1: Pattern/Path of possible input files (separated by |)
    # Input2: Pattern/Path of data files, that shall not be used as input (separated by |)
    # Output: All files that are defined as input in src/config.sh and part of a dataset that is associated to the current project, but not yet input of a workflow. 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_find_files_to_process
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Find files to process
    local pattern=${1:-$(echo "${IN_PREFIX[@]}" | tr ' ' '|')}
    local pattern_excl=${2:-$(echo $(list_plan_outputs "$OUT_PATH") "${IN_FILE['data_generation_script']}" | tr ' ' '|')}
    
    # Check pattern exists
    if [[ -z $pattern ]]
    then
        error_exit "ERROR: Could not detect valid input pattern." \
                   "Pattern either needs to be specified as input argument \$2 or will be derived from IN_PREFIX as specified in src/config.sh." \
                   "Please check find_files_to_process -h for more details."
    fi
    
    unprocessed_in=$(find_unique $(list_dataset_files) $(list_plan_inputs) \
    | grep -E "$pattern")
    
    if [[ -z $pattern_excl ]]
    then
        printf '%s\n' "${unprocessed_in}"
    else
        printf '%s\n' "${unprocessed_in}" | grep -E -v "$pattern_excl"
    fi
    
}


function find_datasets_to_process(){
    
    # Input: json file with the provenance graph stored. Default .renku/provenance.json, which is generated by renku graph generate.
    # Output: All datasets associated to the current project with an input defined in src/config.sh, but not yet input of a workflow. 
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_find_datasets_to_process
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Find datasets to process
    find_files_to_process | cut -f2 -d '/' | sort | uniq
}


function run_command_on_unprocessed(){
    
    # Input1: Command to run on unprocessed files. Needs to be valid for all unprocessed input files.
    # Input2: json file with the provenance graph stored. Default .renku/provenance.json, which is generated by renku graph generate.
    # Output: Command output for all unprocessed files. See find_files_to_process -h.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_run_command_on_unprocessed
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Run command
    local command=$1
    files=($(find_files_to_process))
    for one_file in "${files[@]}"
    do
        $command "$one_file"
    done
}


generate_input_json(){

    # Input1: List of files to write to csv
    # Input2: Output file path
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_generate_input_json
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    # Run command
    local file_list=$*
    
    json_out='{"info_file":[]}'
    
    for file in $file_list
    do
       json_out=$(echo "$json_out" | jq --argjson fil "$(printf '"%s"' "$file")" '.info_file += [$fil]')
    done
   
    echo -e "${json_out}\n"
    
}

##############################################################################################

