#!/usr/bin/env bash

##############################################################################################
#                          Summary Utils Functions                                           #
#                                                                                            #
# Functions to summarize omnibenchmark results, specifically to perform cross repository     #
# queries and identify file sources, KG paths along edges.                                   #
# Most functions are generalizable, but some specifically target omnibenchmark scripts and   #
# are only valid or useful in an omnibenchmark context/repository                            #
#                                                                                            #
# Change History                                                                             #
# 09/07/2021  Almut Lütge         Initialize summary functions and generating setup.         #
#                                                                                            #
# Authors: Almut Lütge                                                                       #
##############################################################################################
##############################################################################################
##############################################################################################

##############################################################################################

## --------------------------------------------------------------------------------------- ##
## -------------------------------- Prerequisites  ----------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/config.sh
} || { 
    echo "Warning, src/config.sh does not exist and may be required for some functions."
}

{ # try
    source src/utils/bash/dataset_utils.sh
} || { 
    echo "Warning, src/utils/check_datasets.sh does not exist and may be required for some functions."
}

# Error function
error_exit()
{
  printf '%s\n' "$@" 1>&2
  exit 1
}


## --------------------------------------------------------------------------------------- ##
## -------------------------------- Global variables  -------------------------------------##
## --------------------------------------------------------------------------------------- ##

{ # try
    source src/utils/global_vars.sh
} || { 
    echo "Warning, src/utils/global_vars.sh does not exist and may be required for some functions."
}


##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Help Documentations  ------------------------------------##
## ---------------------------------------------------------------------------------------- ##


help_get_lineage_from_result()
{
   echo "Function to extract lineage information (dataset, metric, method or parameter) from a result file. Lineage can be one of 'dataset', 'method', 'metric' or 'parameter'"
   echo
   echo "Syntax: get_lineage_from_result RES_FILE_NAME"
   echo
   echo "Arguments:"
   echo "RES_FILE_NAME      Name of a metric result file."
   echo
   echo "Options:"
   echo "h                  Print this Help."
   echo
}


help_generate_json_from_res_files()
{
   echo "Function to generate a summary json with all lineage information from a list of result files."
   echo
   echo "Syntax: generate_json_from_res_files RES_FILE_LIST"
   echo
   echo "Arguments:"
   echo "RES_FILE_LIST     List of result file names to extract lineage informations from"
   echo
   echo "Options:"
   echo "h                 Print this Help."
   echo
}

help_get_lineage_input()
{
   echo "Function to identify source/input files of a list of files within their respective projects."
   echo
   echo "Syntax: get_lineage_input FILE_LIST"
   echo
   echo "Arguments:"
   echo "FILE_LIST         List of file pathes to extract lineage information for"
   echo
   echo "Options:"
   echo "h                 Print this Help."
   echo
}

##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ------------------------------- Helper functions/snippets -------------------------------##
## ---------------------------------------------------------------------------------------- ##

function graphql_query () {
    
    # Input1: Path to file to query
    # Input2: Path to project to perform query in
    # Output: Nodes and edges for a specific graphql endpoint
    
    local file_path=$1
    local project_path=$2

    curl -X POST -v -H "Content-Type: application/json" "$GRAPHQL_URL" -d '{ "query": "{ lineage(projectPath: \"'"${project_path}"'\", filePath: \"'"${file_path}"'\") { nodes { id label } edges { source target } } }"}'
    
}


function get_upstream_project_by_file () {
    
    # Input1: Path to file to query
    # Input2: Path to project to perform query in
    # Output: Nodes and edges for a specific graphql endpoint
    
    local in_file=$1
    local link_url="${GIT_URL}/"
   
    project=$(renku dataset ls-files --format json-ld | jq -r '.[] | select( ."http://www.w3.org/ns/prov#atLocation" | contains("'"$in_file"'")) | ."http://schema.org/isBasedOn" | ."http://schema.org/url"' | sed 's,'"${link_url}"',,' | sed 's,.git,,')
    
    # check upstream project exists 
    if [[ "$project" =~ null || -z $project ]]
    then
        error_exit "ERROR: No upstream project for file $in_file found." \
                   "Make sure the file was imported from another renku project's dataset."
    fi
    
    echo "$project"
    
}



##############################################################################################

## ---------------------------------------------------------------------------------------- ##
## ---------------------------------  Perform queries  -------------------------------------##
## ---------------------------------------------------------------------------------------- ##


function get_lineage_input () {

    # Input1: Source files to find the lineage for
    # Output: Source files used generate all input files
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_lineage_input
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   local input_files=$*
     
    # Find dataset ids of input files
    sou_ids=()
   
    for in_file in $input_files
    do
        # check file to be part of a dataset (be in data/)
        sou_dir=$(dirname "$in_file")
        
        if [[ ! $sou_dir == *"data/"* ]]
        then
             error_exit "ERROR: file path is not in the data folder." \
                        "Please make sure that the input files are placed in data/."
        fi
        
        sou_data="${sou_dir//data\//}"
        sou_ids+=("$(curl -s "${KG_URL}/datasets?query=$sou_data" | jq -r '.[] | .identifier')")
    done
    
    source_ids=$(printf '%s\n' "${sou_ids[@]}" | sort | uniq)
    
    # Find original IDs
    in_ids=()
    for id in ${source_ids}
    do
       in_ids+=( "$(get_sameAs_data_id "$id")" )  
    done
    
    origin_ids=$(check_import_from_target -a 2 "${in_ids[@]}")
    
    # Find source files
    source_files=()
    for id in ${origin_ids[@]}
    do
        project_path=$(curl -s "${KG_URL}/datasets/$id" | jq -r '.'\"project\"' | .'\"path\")
        id_files=$(dataset_files_by_id "$id")
        file=$(printf '%s\n' ${input_files[@]} ${id_files[@]} | sort | uniq -d | head -1)
        
        source_files+=("$(graphql_query "$file" "$project_path" | \
                            jq -r '.data | .lineage | .edges[] | select(.target | contains('\"/runs/\"')) | .source')")  
    done
   
   echo "${source_files[@]}"
}




function get_dataset_from_result () {

    # Input: Result file from metric results
    # Output: Name of the dataset related to this result.
    
    # Display help message
    local OPTIND 
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_lineage_from_result
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
    local res_file=$1
    local link_url="${GIT_URL}/"
    local res_project
   
    res_project=$(get_upstream_project_by_file "$res_file")
   
    input_files=$(graphql_query "$res_file" "$res_project" | \
                   jq -r '.data | .lineage | .edges[] | select(.target | contains('\"/runs/\"')) | .source' | \
                   grep "data/")
    
   # Check correct input file
   if [[ -z ${input_files} ]]
   then
       error_exit "ERROR: No source files to generate the input file found." \
                  "Please check the input to be a file path to an existing metric result."
   fi
    
    input_filtered=$(printf '%s\n' "$input_files" | grep -E -v "meta_")
    
    method_input=$(get_lineage_input "$input_filtered")
    method_filtered=$(printf '%s\n' "$method_input" | grep -E "data/" | grep -E -v "meta_")
    
    process_input=$(get_lineage_input "$method_filtered")
    process_filtered=$(printf '%s\n' "$process_input" | grep -E "data/" | grep -E -v "meta_")
    
    # Check correct input file
   if [[ -z ${process_filtered} ]]
   then
       error_exit "ERROR: No source files to generate the input file found." \
                  "Please check the input to be a file path to an existing metric result."
   fi
        
    # Get dataset name of input files
    datasets=()
   
    for pr_file in $process_filtered
    do
        datasets+=( "$(dirname "$pr_file" | sed 's/data\///')")
    done
    
    echo "${datasets[@]}"   
}


function get_method_from_result () {

    # Input: Result file from metric results
    # Output: Name of the method related to this result.
    
    local OPTIND prefix="_-_placeholder_-_"
    
    # Display help message
    while getopts ":p:h" option; do
        case $option in
            h) 
                help_get_lineage_from_result
                return 1;;
            p)
                prefix=${OPTARG};;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   local res_file=$1
   local link_url="https://renkulab.io/gitlab/"
   local res_project
   
   res_project=$(get_upstream_project_by_file "$res_file")
   
   input_files=$(graphql_query "$res_file" "$res_project" | \
                   jq -r '.data | .lineage | .edges[] | select(.target | contains('\"/runs/\"')) | .source' | \
                   grep "data/")
    
   # Check correct input file
   if [[ -z ${input_files} ]]
   then
       error_exit "ERROR: No source files to generate the input file found." \
                  "Please check the input to be a file path to an existing metric result."
   fi
   
   input_filtered=$(printf '%s\n' "$input_files" | grep -E -v "meta_")
   
   # Get method name of input files
    methods=()
   
    for in_file in $input_filtered
    do
        methods+=("$(dirname "$in_file" | sed 's/data\///' | sed 's/'"${prefix}"'//g')")
    done
    
    # Check ouput exists
    if [[ -z ${methods[@]} ]]
    then
        error_exit "ERROR: No method name found." \
                  "Please check the input to be a file path to an existing metric result."
    fi
    
    echo "${methods[@]}"
       
}

function get_metric_from_result () {

    # Input: Result file from metric results
    # Output: Name of the metric related to this result.
    
    local OPTIND prefix="_-_placeholder_-_"
    
    # Display help message
    while getopts ":p:h" option; do
        case $option in
            h) 
                help_get_lineage_from_result
                return 1;;
            p)
                prefix=${OPTARG};;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   local res_file=$1
   
   metric=$(dirname "$res_file" | sed 's/data\///' | sed 's/'"${prefix}"'//g')
   
   # Check ouput exists
   if [[ -z ${metric} ]]
   then
       error_exit "ERROR: No metric name found." \
                  "Please check the input to be a file path to an existing metric result."
   fi

   echo "$metric"
   
}


function get_parameter_from_result () {

    # Input: Result file from metric results
    # Output: Name of the parameter related to this result.
    
    # Display help message
    local OPTIND
    while getopts ":h" option; do
        case $option in
            h) 
                help_get_lineage_from_result
                return 1;;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   local res_file=$1
   
   filename=$(basename -- "$res_file")
   filename="${filename%.*}"
   
   
   #method=$(echo $filename | sed 's/.*res_//g' | sed 's/_.*//g')
   #parameter=$(echo $filename | sed 's/.*'"$method"_'//g' | sed 's/__/ /g' | sed 's/_/:/g')
    
   
   ## ONLY for prototype until parameter and cross repo queries are enabled
   length_harmony=$(echo "$filename" | grep "harmony" | wc -w)
   
   case $length_harmony in
            0) 
                parameter="order:auto hvg:NULL d:20 k:30";;
            1)
                parameter="d:20 k:30 lambda:1";;
            esac
      
   echo "$parameter"
   
}

function generate_json_from_res_files () {
    # Input: List of result files from metric outputs
    # Output: json file with the result files run informations as metric name, input dataset, method name and run parameter
    
    local OPTIND prefix="_-_placeholder_-_"
    
    # Display help message
    while getopts ":p:h" option; do
        case "${option}" in
            h) 
                help_generate_json_from_res_files
                return 1;;
            p)
                prefix="${OPTARG}";;
            \?) 
                error_exit "Error: Invalid option";;
            esac
    done
    shift $((OPTIND-1))
    
   local res_files=$*
   
   local json_out='{}'
   
   for file in $res_files
   do
       val='{"dataset": [], "method":[],  "parameter":{}, "metric":[]}'
       json_out=$(jq -n --arg key "$file" --argjson out "$json_out" --argjson val "$val" '$out + {($key): $val}')
       json_out=$(echo "$json_out" | jq --argjson data "$(printf '"%s"' "$(get_dataset_from_result "$file")")" --arg fil "$file" '.[$fil] += {dataset: [$data]}')
       json_out=$(echo "$json_out" | jq --argjson method "$(printf '"%s"' "$(get_method_from_result -p "$prefix" "$file")")" --arg fil "$file" '.[$fil] += {method: [$method]}')
       json_out=$(echo "$json_out" | jq --argjson metric "$(printf '"%s"' "$(get_metric_from_result -p "$prefix" "$file")")" --arg fil "$file" '.[$fil] += {metric: [$metric]}')
       
       parameter=$(get_parameter_from_result "$file")
       for param in $parameter
       do
           param_name=$(echo "$param" | sed 's/:.*//g')
           param_value=$(echo "$param" | sed 's/.*://g')
           json_out=$(echo "$json_out" | jq --argjson val "$(printf '"%s"' "$param_value")" --arg key "$param_name" --arg fil "$file" '.[$fil].parameter += {($key): $val}')
       done
   done
   
   echo -e "${json_out}\n"
}

